# coding=utf-8
# Copyright 2024 Pandora Media, LLC.
#
# Licensed under the GNU GPL License, Version 3.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gnu.org/licenses/gpl-3.0.en.html
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python standard library imports
# None.

# Third party imports
from scooch import Configurable

# Local imports
# None.


class Extractor(Configurable):
    """
    Base class for classes that are responsible for extracting data
    from mule.features.Feature classes.
    """

    # 
    # Methods
    #
    def extract_range(self, feature, start_index, end_index):
        """
        Extracts data over a given index range from a single feature.

        Args:
            feature: mule.feature.Feature - A feature to extract data from.

            start_index: int - The first index (inclusive) at which to return data.

            end_index: int - The last index (exclusive) at which to return data.

        Return:
            numpy.ndarray - The extracted feature data. Features on first axis, time on
            second axis.
        """
        raise NotImplementedError(f"The {self.__class__.__name__} class has no `extract_range` method.")

    def extract_batch(self, features, indices):
        """
        Extracts a batch of features from potentially multiple features, each potentially
        at distinct indices.

        Args:
            features: list(mule.features.Feature) - A list of features from which to extract 
            data from.

            indices: list(int) - A list of indices, the same size as `features`. Each element
            provides an index at which to extract data from the coressponding element in the
            `features` argument.

        Return:
            np.ndarray - A batch of features, with features on the batch dimension on the first
            axis and feature data on the remaining axes.
        """
        raise NotImplementedError(f"The {self.__class__.__name__} class has no `extract_batch` method.")



# coding=utf-8
# Copyright 2024 Pandora Media, LLC.
#
# Licensed under the GNU GPL License, Version 3.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gnu.org/licenses/gpl-3.0.en.html
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Local imports
from ..analysis import Analysis

# Third party imports
import click
from scooch import Config
from . import options

# Python standard library imports
# None.


@click.group()
def main():
    pass


@main.command("analyze")
@options.config
@options.input_file
@options.output_file
def run(config, input_file, output_file):
    """
    Constructs the analysis pipeline, adds options to it, and runs it.
    """
    cfg = Config(config)
    analysis = Analysis(cfg)
    feat = analysis.analyze(input_file)
    feat.save(output_file)


if __name__=='__main__':
    main()



# coding=utf-8
# Copyright 2021 Pandora Media, LLC.
#
# Licensed under the GNU GPL License, Version 3.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gnu.org/licenses/gpl-3.0.en.html
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python standard library imports
# None.

# Third party imports
import tensorflow as tf

# Local imports
# None.


class StochDepth(tf.keras.layers.Layer):
    """
    Stochastic Depth layer. Implements Stochastic Depth as described in
    [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382), 
    to randomly drop residual branches in residual architectures.
    """

    def __init__(self, survival_probability: float = 0.5, scale_during_test: bool = False, **kwargs):
        """
        **Constructor**

        Args:
            survival_probability: <float> - The probability of the residual branch being kept.

            scale_during_test: <bool> - Whether to apply scaling to residual branch activations during test 
            time, after training, to compensate for the fact that branches are ranodmly dropped during training.
        """
        super().__init__(**kwargs)

        self.survival_probability = survival_probability
        self.scale_during_test = scale_during_test

    def call(self, x, training=None):
        """
        Apply the layer to the inputs, thereby combining the residual and skip layers via addition,
        with a random probability of dropping the resiudal component from the addition.

        Args:
            inputs: <list(tf.Tensor)> - A two element list containing a `shortcut` and `residual`
            tensor. Both tensors must be of the same shape.

            training: <tf.Tensor> - A boolean tensor defining whether it is train or test time. 

        Return:
            <tf.Tensor> - The tensor of the combined residual and skip connections.
        """
        if not isinstance(x, list) or len(x) != 2:
            raise ValueError("input must be a list of length 2.")

        shortcut, residual = x

        # Random bernoulli variable indicating whether the branch should be kept or not or not
        b_l = tf.keras.backend.random_bernoulli(
            tf.shape(x[1])[:1], p=self.survival_probability, dtype=self._compute_dtype_object
        )
        # NOTE [matt.c.mccallum 03.30.22]: Assume that this is being used on a 2D convolutional layer here.
        b_l = b_l[:, tf.newaxis, tf.newaxis, tf.newaxis]

        def _call_train():
            return shortcut + b_l * residual

        def _call_test_scale():
            return shortcut + self.survival_probability * residual

        def _call_test_no_scale():
            return shortcut + residual

        if self.scale_during_test:
            return tf.keras.backend.in_train_phase(
                _call_train, _call_test_scale, training=training
            )
        else:
            return tf.keras.backend.in_train_phase(
                _call_train, _call_test_no_scale, training=training
            )

    def compute_output_shape(self, input_shapes):
        """
        The size of the output, identical to both input.

        Args:
            input_shapes: <list(tf.Tensor)> - The sizes of each of the inputs.

        Return:
            <tf.Tensor> - The size of the output.
        """
        return input_shapes[0]

    def get_config(self):
        """
        Get the model parameters so that it may be reconstructed after serialization.

        Return:
            <dict(str, obj)> - Just the parameters of the constructor which will be
            automatically applied at deserialization.
        """
        base_config = super().get_config()

        config = {
            "survival_probability": self.survival_probability,
            "scale_during_test": self.scale_during_test
        }

        return {**base_config, **config}



# coding=utf-8
# Copyright 2021 Pandora Media, LLC.
#
# Licensed under the GNU GPL License, Version 3.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gnu.org/licenses/gpl-3.0.en.html
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Local imports
# None.

# Third party imports
from scooch import Configurable
import tensorflow as tf

# Python standard library imports
# None.


class Loss(Configurable):
    """
    A base class for keras or custome loss functions.
    """

    def init(self):
        """
        Initialize the Tensorflow Loss class that this class inherits from. This is separated from the class's
        constructor, so that it can be initialized at any time, under the desired Tensorflow scopes (e.g., a
        multi-GPU distribution scope), rather than only at construction time.
        """
        self.initialize_base()

    @classmethod
    def from_config(cls, config):
        if 'cfg' not in config.keys():
            cfg = {cls._NAME: None}
        else:
            cfg = config.pop('cfg')
        inst = cls(cfg=cfg, **config)
        inst.init()
        return inst



# coding=utf-8
# Copyright 2021 Pandora Media, LLC.
#
# Licensed under the GNU GPL License, Version 3.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gnu.org/licenses/gpl-3.0.en.html
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python standard library imports
import tempfile

# Third party imports
from scooch import Configurable
import numpy as np

# Local imports
# None.


class Feature(Configurable):
    """
    The base class for all feature types.
    """

    def __del__(self):
        """
        **Destructor**
        """
        # NOTE [matt.c.mccallum 09.20.22]: This destructor is not guaranteed to be called on interpreter exit..
        if hasattr(self, '_data_file'):
            self._data_file.close()

    def add_data(self, data):
        """
        Adds data to extend the object's current data via concatenation along the time axis.
        This is useful for populating data in chunks, where populating it all at once would
        cause excessive memory usage.

        Arg:
            data: np.ndarray - The data to be appended to the object's memmapped numpy array.
        """
        if not hasattr(self, '_data_file'):
            self._data_file = tempfile.NamedTemporaryFile(mode='w+b')

        if not hasattr(self, '_data') or self._data is None:
            original_data_size = 0
        else:
            original_data_size = self._data.shape[1]
        final_size = original_data_size + data.shape[1]

        filename = self._data_file.name

        # NOTE [matt.c.mccallum 06.22.20]: By using the np.memmap method instead of numpy.lib.format.open_memmap, 
        #                                  we can incrementally change the size of the array on disk, but we aren't
        #                                  able to save the shape and data type metadata.
        self._data = np.memmap(filename, dtype='float32', mode='r+',shape=(data.shape[0],final_size), order='F')
        self._data[:,original_data_size:] = data

    def _extract(self, source, length):
        """
        Extracts feature data file or other feature a given time-chunk.

        Args:
            source_feature: mule.features.Feature - The feature to transform.

            start_time: int - The index in the input at which to start extracting / transforming.

            chunk_size: int - The length of the chunk following the `start_time` to extract
            the feature from.
        """
        raise NotImplementedError(f"The {self.__name__.__class__} has no feature extraction method")

    def save(self, path):
        """
        Save the feature data blob to disk.

        Args:
            path: str - The path to save the data to.
        """
        np.save(path, self.data)

    def clear(self):
        """
        Clears any previously analyzed feature data, ready for a new analysis.
        """
        self._data = None
        if hasattr(self, '_data_file'):
            self._data_file.close()
            del self._data_file

    @property
    def data(self):
        """
        The feature data blob itself.
        """
        return self._data



# coding=utf-8
# Copyright 2024 Pandora Media, LLC.
#
# Licensed under the GNU GPL License, Version 3.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gnu.org/licenses/gpl-3.0.en.html
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python standard library imports
# None.

# Third party imports
# None.

# Local imports
from .extractor import Extractor


class BlockExtractor(Extractor):
    """
    Extracts a single block of data from the provided feature, over the provided time range.
    """

    # 
    # Methods
    #
    def extract_range(self, feature, start_index, end_index):
        """
        Extract all feature indices over a given index range.

        Args:
            feature: mule.feature.Feature - A feature to extract data from.

            start_index: int - The first index (inclusive) at which to return data.

            end_index: int - The last index (exclusive) at which to return data.

        Return:
            numpy.ndarray - The extracted feature data. Features on first axis, time on
            second axis.
        """
        return feature.data[:, start_index:end_index]



# coding=utf-8
# Copyright 2021 Pandora Media, LLC.
#
# Licensed under the GNU GPL License, Version 3.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gnu.org/licenses/gpl-3.0.en.html
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python standard library imports
import itertools

# Third party imports
import tensorflow as tf
import numpy as np
from scooch import Param

# Local imports
from .layers import WeightStandardization
from .layers import ScalarMultiply
from .layers import StochDepth
from .model import Model


def _scaled_activation(activation_name):
    """
    Apply a scaled activation function according to [1, 2].

    Args:
        activation_name: str - The name of the scaled activation function to apply.

        input: tf.Tensor - The tensor to apply the activation to.

    Return:
        tf.Tensor - The input with the activation applied.

    References:
    
        [1] Arpit, Devansh, Yingbo Zhou, Bhargava Kota, and Venu Govindaraju. "Normalization propagation: A 
            parametric technique for removing internal covariate shift in deep networks." In International 
            Conference on Machine Learning, pp. 1168-1176. PMLR, 2016.

        [2] Brock, A., De, S., & Smith, S. L. (2021). Characterizing signal propagation to close the performance 
            gap in unnormalized resnets. arXiv preprint arXiv:2101.08692.
    """
    activations = {
        'gelu': lambda x: tf.nn.gelu(x) * 1.7015043497085571,
        # NOTE [matt.c.mccallum 03.22.22]: The scalar below is taken directly from DeepMind reference code,
        #      it is similar but not equal to the value suggested in the literature of: np.sqrt(2.0/(1-(1/np.pi)))
        'relu': lambda x: tf.nn.relu(x) * 1.7139588594436646 
    }
    return activations[activation_name]


class SfNfNetF0(Model):
    """
    This model implements the SF-NFNet-F0 architecture described in [1].
    It consists of two series' of NFNet-F0 stages [2], one with high time
    resolution and low channel count, another with low time resolution and
    high channel count, based on the effectiveness of this for audio, previously
    demonstrated in [3].

    Much of this code (including the NFNet) blocks was adapted from the architecture
    written in Haiku by DeepMind:

        https://github.com/deepmind/deepmind-research/blob/1642ae3499c8d1135ec6fe620a68911091dd25ef/nfnets/nfnet.py

    Reference code for specfic NFNet blocks such as Squeeze and Excitation blocks, 
    Weight Standardized Convolutions and Stochastic Depth layers can be found here:

        https://github.com/deepmind/deepmind-research/blob/1642ae3499c8d1135ec6fe620a68911091dd25ef/nfnets/base.py

    References:

        [1] Wang, Luyu, Pauline Luc, Yan Wu, Adria Recasens, Lucas Smaira, Andrew Brock, 
        Andrew Jaegle et al. "Towards Learning Universal Audio Representations." arXiv 
        preprint arXiv:2111.12124 (2021).

        [2] Brock, Andy, Soham De, Samuel L. Smith, and Karen Simonyan. "High-performance 
        large-scale image recognition without normalization." In International Conference 
        on Machine Learning, pp. 1059-1071. PMLR, 2021.

        [3] Kazakos, Evangelos, Arsha Nagrani, Andrew Zisserman, and Dima Damen. 
        "Slow-fast auditory streams for audio recognition." In ICASSP 2021-2021 IEEE 
        International Conference on Acoustics, Speech and Signal Processing (ICASSP), 
        pp. 855-859. IEEE, 2021.
    """

    # TODO [matt.c.mccallum 03.18.22]: Currently this network is a huge class with several private
    #      `_make_<module_name>_module` and `_apply_<module_name>_module` methods. Each of these pairs
    #      of layers forming modules should be should in-turn, also be layers.

    # 
    # SCOOCH Configuration
    #
    _projector_activation = Param(
        str,
        default='relu',
        doc="Activation function used in any projection layers prior to the network output."
    )
    _projector_layers = Param(
        list,
        doc=")> - Number of nuerons in each dense projector layers between the convolutional layers, and the output. Also defines the number of dense layers there. Can be an empty list."
    )
    _include_fc = Param(
        bool,
        default=True,
        doc="Include a dense layer between the convolutional layers, projection, and loss function"
    )
    _alpha = Param(
        float,
        default=0.2,
        doc="The variance scaling factor to be applied at the end of each residual connection"
    )
    _scaled_activation_type = Param(
        str,
        default='gelu',
        doc="The activation function to use when applying scaled activations"
    )
    _f_value = Param(
        int,
        default=0,
        doc="The 'F' value for the NFNet architecture. An F-Value > 0 multiplies the number of NFNet blocks in each stage by (F+1)"
    )

    # 
    # Methods
    #
    def _make_model(self):
        """
        Constructs the model object.

        Return:
            tf.keras.Model - A keras model object.
        """
        inputs = [tf.keras.Input(shape=self._input_shapes[0][1:])]

        self._make_layers(self._output_shapes[0][1])

        # Add activation for classification
        self._output_layers += [tf.keras.layers.Activation(tf.keras.activations.sigmoid)]

        outputs = [self._apply_layers(inp) for inp in inputs]

        model = self._MODEL_CLASS(inputs, outputs)

        return model

    def get_conv(self, channels, kernel, strides, groups, activation=None):
        """
        Creates the variance preserving weight standardized convolutional layer as described in [1].
        This is provided as a one or two element list, depending on whether an activation is provided.
        If an activation function is provided, it will be implemented as a separate tf.keras.layers.Activation
        layer.

        Args:

            channels: <int> - The number of channels at the output of the layer.

            kernel: <list(int)> - The dimensions of the 2D convolutional kernel.

            strides: <list(int)> - The convolutional stride across each direction in the 2D features.

            groups: <int> - The number of channel groups the input will be split into for a grouped 
            convolution operation.

            activation: <None or function or str> - The activation function applied after the convolution,
            This should be implemented according to the variance scaling rules specified in [2].

        References:

            [1] Brock, Andy, Soham De, Samuel L. Smith, and Karen Simonyan. "High-performance 
            large-scale image recognition without normalization." In International Conference 
            on Machine Learning, pp. 1059-1071. PMLR, 2021.

            [2] Brock, A., De, S., & Smith, S. L. (2021). Characterizing signal propagation to close the performance 
            gap in unnormalized resnets. arXiv preprint arXiv:2101.08692.
        """

        # NOTE [matt.c.mccallum 04.01.22]: The DeepMind reference code written in Haiku applies Variance scaling initialization
        #      to the unnormalized kernel, they then normalize the kernel. So the initial value of the kernel regardless is
        #      1.0 averaged across the fan_in dimension, regardless of the variance scaling initialization. The weight scaled 
        #      normalization we have implemented in Keras takes the same approach.

        layers = [
            WeightStandardization(tf.keras.layers.Conv2D(
                filters=channels,
                kernel_size=kernel,
                strides=strides,
                padding="same",
                groups=groups,
                kernel_initializer=tf.keras.initializers.VarianceScaling(
                    scale=1.0,
                    mode='fan_in',
                    distribution='untruncated_normal'
                )
            ))
        ]

        if activation is not None:
            return layers + [tf.keras.layers.Activation(activation)]
        else:
            return layers

    def _make_layers(self, output_shape):
        """
        Makes the layers in the model and saves them as Model class attributes,
        without applying the layers to any variables.
        """

        #
        # Make Stems
        #
        self._slow_stem = self._make_stem_module(
                kernels=[
                    [3, 1],
                    [3, 1],
                    [3, 1],
                    [3, 3]
                ],
                channels=[16, 32, 64, 128],
                strides=[
                    [2, 8], # Integrated data striding layer into first convolution
                    [1, 1],
                    [1, 1],
                    [2, 2]
                ]
            )

        self._fast_stem = self._make_stem_module(
                kernels=[
                    [3, 3],
                    [3, 3],
                    [3, 3],
                    [3, 3]
                ],
                channels=[2, 4, 8, 16],
                strides=[
                    [2, 2], # Integrated data striding layer into first convolution
                    [1, 1],
                    [1, 1],
                    [2, 2]
                ]
            )

        #
        # Initialize parameters for NFNet Blocks
        #
        self._nfnet_stage_depths = [x*(self._f_value+1) for x in (1,2,6,3)] # Computes the number of NFNet blocks in each stage based on the provided F-Value
        cumulative_stage_depths = np.concatenate(([0],np.cumsum(self._nfnet_stage_depths)))
        self._stoch_depth_survival_probs = 0.1*np.arange(cumulative_stage_depths[-1])/(cumulative_stage_depths[-1])
        self._stoch_depth_survival_probs = [
            self._stoch_depth_survival_probs[st:end] for st, end in zip(cumulative_stage_depths[:-1], cumulative_stage_depths[1:])
        ]
        self._stage_expected_vars = [1.0] + [(1.0+self._alpha**2)**0.5]*3 # In the reference code, an initial beta of 1.0 is applied at the first stage, and betas after a single update at the latter stages.
        self._stage_downsamples = [1] + [2]*3

        #
        # Construct NFNet stages for the slow path
        #
        slow_nfnet_kernels = [
            [[1, 1],[1, 3],[3, 1],[1, 1]],
            [[1, 1],[1, 3],[3, 1],[1, 1]],
            [[1, 1],[1, 3],[3, 1],[1, 1]],
            [[1, 1],[1, 3],[3, 1],[1, 1]]
        ]
        slow_nfnet_input_sizes = [256, 512, 1024, 2560]
        slow_nfnet_output_sizes = [256, 512, 1536, 1536]

        self._slow_layers = [
            self._make_nfnet_stage(
                kernels=k,
                freq_downsample=f,
                input_channels=i,
                output_channels=o, 
                group_size=128, 
                alpha=self._alpha, 
                input_expected_var=e,
                stoch_depths=s,
                num_blocks=n
            ) for k, f, i, o, e, s, n in zip(
                slow_nfnet_kernels,
                self._stage_downsamples,
                slow_nfnet_input_sizes,
                slow_nfnet_output_sizes,
                self._stage_expected_vars,
                self._stoch_depth_survival_probs,
                self._nfnet_stage_depths
            )
        ]

        #
        # Construct NFNet stages for the fast path
        #
        fast_nfnet_kernels = [
            [[1, 1],[1, 3],[3, 1],[1, 1]],
            [[1, 1],[1, 3],[3, 1],[1, 1]],
            [[1, 1],[1, 3],[3, 1],[1, 1]],
            [[1, 1],[1, 3],[3, 1],[1, 1]]
        ]
        fast_nfnet_input_sizes = [16, 32, 64, 192]
        fast_nfnet_output_sizes = [32, 64, 192, 192]

        self._fast_layers = [
            self._make_nfnet_stage(
                kernels=k,
                freq_downsample=f,
                input_channels=i,
                output_channels=o, 
                group_size=16, 
                alpha=self._alpha, 
                input_expected_var=e,
                stoch_depths=s,
                num_blocks=n
            ) for k, f, i, o, e, s, n in zip(
                fast_nfnet_kernels,
                self._stage_downsamples,
                fast_nfnet_input_sizes,
                fast_nfnet_output_sizes,
                self._stage_expected_vars,
                self._stoch_depth_survival_probs,
                self._nfnet_stage_depths
            )
        ]

        #
        # Construct fast-to-slow fusion layers
        #
        # NOTE [matt.c.mccallum 04.01.22]: Conversation with the authors of the SF-NFNet-F0 stated that a 1x1 conv layer
        #      in fast to slow fusion expands the feature dimension by [2,4,8,16] times. They also mentioned that the inputs
        #      to each of the NFNet stages had [256,512,1024,2560] channels. These two facts do not add up as the expansion by
        #      the stated factors causes the stages to have inputs of [256,512,1024,4608] channels. We go with the stated factors
        #      of [2,4,8,16] because it brings the model's total parameter count closer to the stated value in the paper (62M 
        #      parameters here, 63M parameters mentioned in the paper).
        self._fusion_layers = [
            self._make_fast_to_slow_fusion(time_kernel_length=7, time_stride=4, input_channels=32, output_channels=128),
            self._make_fast_to_slow_fusion(time_kernel_length=7, time_stride=4, input_channels=32, output_channels=256),
            self._make_fast_to_slow_fusion(time_kernel_length=7, time_stride=4, input_channels=64, output_channels=512),
            self._make_fast_to_slow_fusion(time_kernel_length=7, time_stride=4, input_channels=192, output_channels=3072)
        ]

        #
        # Construct summarization and aggregation layers at the output
        #
        self._output_layers = [
            tf.keras.layers.GlobalAveragePooling2D(),   # Slow path
            tf.keras.layers.GlobalAveragePooling2D(),   # Fast path
            tf.keras.layers.Concatenate(),
            tf.keras.layers.Activation(_scaled_activation(self._scaled_activation_type))
        ]

        #
        # Any projector / dense net to be attached to the residual network output pre-loss function
        #
        # TODO [matt.c.mccallum 03.22.22]: It is not clear from the literature and reference code wherther projectors use
        #      scaled or unscaled activations. For now these are unscaled, but we should clarify this as more informaton
        #      becomes available.
        # TODO [matt.c.mccallum 04.08.22]: It is also not clear whether these projectors have batch normalization layers
        #      interspersed, as they do in Google's SimCLR work. Because this is a normalizer free network, we leave thes
        #      batch norm layers out for now.
        self._projectors = [
            tf.keras.layers.Dense(num_nuerons, activation=self._projector_activation, name=f'projector_{idx}') 
            for idx, num_nuerons in enumerate(self._projector_layers)
        ]

        # TODO [matt.c.mccallum 11.22.22]: Make the use of bias configurable here.
        if self._include_fc:
            self._projectors += [tf.keras.layers.Dense(
                output_shape, 
                activation=None,
                kernel_initializer=tf.random_normal_initializer(mean=0.0, stddev=0.01), 
                name='fc',
                use_bias=False # NOTE [matt.c.mccallum 04.08.22]: This shouldn't have a bias according to the SimCLR reference code here: https://github.com/google-research/simclr/blob/master/model_util.py#L141-L177
            )]

    def _apply_layers(self, input_tensor):
        """
        Take the models layers that have been constructed and apply them to an input tensor.

        Args:
            input_tensor: <tf.Tensor> - The input tensor that the model's layers will be applied to.

        Return:
            <tf.Tensor> - The output, after applying the model's layers.
        """

        # Apply stem modules
        slow = self._apply_stem(self._slow_stem, input_tensor)
        fast = self._apply_stem(self._fast_stem, input_tensor)

        # For each nfnet_transition module
        for fuse, slw_lyr, fst_lyr in zip(self._fusion_layers, self._slow_layers, self._fast_layers):
            slow = self._apply_fast_to_slow_fusion(fuse, slow, fast)
            slow = self._apply_nfnet_stage(slw_lyr, slow)
            fast = self._apply_nfnet_stage(fst_lyr, fast)

        # Apply global average pool and concat
        slow_out = self._output_layers[0](slow)
        fast_out = self._output_layers[1](fast)
        output = self._output_layers[2]([slow_out, fast_out])
        output = self._output_layers[3](output)

        # Apply projector layers
        for proj in self._projectors:
            output = proj(output)

        return output

    def _make_stem_module(self, kernels, channels, strides):
        """
        Create the stem module. This is a series of convolutional layers that are applied on
        the input, prior to any residual stages.

        Args:
            kernels: <list(list(int))> - Size of the 2D convolutional kernels, at each stem layer.

            channels: <list(int)> - Number of channels at the output of each convolutional layer.

            strides: <list(list(int))> - The stride of each 2D convolutional kernel, at each stem layer.

        Return:
            <list(tf.keras.layers.Layer)> - The constructed layers for the stem module.
        """
        activations = [_scaled_activation(self._scaled_activation_type)]*(len(kernels)-1) + [None]
        layers = [
            self.get_conv(
                channels=c,
                kernel=k,
                strides=s,
                groups=1,   # No groups in stem layers
                activation=a
            ) for c, k, s, a in zip(channels, kernels, strides, activations)
        ]
        return list(itertools.chain(*layers))

    def _apply_stem(self, stem_module, input):
        """
        Applies a stem module to an input.

        Args:
            stem_module: <list(tf.keras.layers.Layer)> - Layers constructed by the `_make_stem_module` method.

            input: tf.Tensor - The input tensor to apply the layers to.

        Return:
            <tf.Tensor> - The output, after applying the stem module layers.
        """
        for lyr in stem_module:
            input = lyr(input)
        return input

    def _make_nfnet_stage(self, kernels, freq_downsample, input_channels, output_channels, group_size, alpha, input_expected_var, stoch_depths, num_blocks):
        """
        Constructs an NFNet stage, made up of multiple NFNet blocks.

        Args:
            kernels: <list(list(int))> - The size of each of the 2D convolutional kernels, identical 
            in each NFNet block.

            freq_downsample: <int> - The factor to downsample the frequency dimension by using the 
            second convolutional layer of the first NFNet block.

            input_channels: <int> - The number of channels at the input to the first NFNet block 
            in this stage.

            output_channels: <int> - The number of channels at the output of every NFNet block in 
            this stage.

            group_size: <int> - The number of inputs to include in each group in the grouped 
            convolutional stages.

            alpha: <float> - The alpha value, used to scale tensor variances for initialization, 
            to ensure unit variance between each block.

            input_expected_var: <float> - The expected variance at the input to the first NFNet 
            block of this stage. Important for scaling for initialization.

            stoch_depths: <list(float)> - The stochastic depth skip probability for each of the 
            NFNet blocks in this stage.

            num_blocks: <int> - The number of NFNet blocks to include in this stage.

        Return:
            list(list(list(tf.keras.layers.Layer)))) - All layers in the NFNet stage. Each item
            in the top level list as an NFNet block. Each item in the second level list includes
            the components of that NFNet stage separated into input layers, residual layers, skip
            layers, and output layers.
        """
        # NFNet transition block first
        blocks = [self._make_nfnet_block(
            kernels, 
            freq_downsample, 
            input_channels, 
            output_channels,
            group_size, 
            alpha,
            1.0/input_expected_var,
            float(stoch_depths[0]),
            is_transition=True
        )]

        # NFNet non-transition blocks
        expected_std = (input_expected_var**2.0 + alpha**2.0)**0.5
        for idx in range(1,num_blocks):
            blocks += [self._make_nfnet_block(
                kernels,
                1, 
                output_channels, 
                output_channels,
                group_size, 
                alpha, 
                1.0/expected_std,
                float(stoch_depths[idx])
            )]
            expected_std = (expected_std**2.0 + alpha**2.0)**0.5
        return blocks

    def _apply_nfnet_stage(self, blocks, input):
        """
        Applies all blocks in an NFNet stage to an input tensor.

        Args:
            blocks: list(list(list(tf.Tensor))) - All layers comprising the NFNet stage, separated into
            NFNet blocks, and in turn into NFBlock components (i.e., input, residual, skip, output).

        Return:
            <tf.Tensor> - The tensor after applying the NFNet stage layers.
        """
        output = input
        for block in blocks:
            output = self._apply_nfnet_block(block, output)
        return output
        
    def _make_nfnet_block(self, kernels, freq_downsample, input_channels, output_channels, group_size, alpha, beta, stoch_depth, is_transition=False):
        """
        Make a single NFNet block, comprised of input layers, residual layers, skip layers and output layers.

        Args:
            kernels: <list(list(int))> - The size of each of the 2D convolutional kernels in this block.

            freq_downsample: <int> - The factor to downsample the frequency dimension by increasing the
            stride of the second convolutional layer.

            input_channels: <int> - The number of channels at the input to this NFNet block.

            output_channels: <int> - The number of channels at the output of this NFNet block.

            group_size: <int> - The number of input channels to include in each group in the grouped 
            convolutional stages.

            alpha: <float> - The alpha value, applied as a scalar multiplier at the input. Used to 
            scale tensor variances for initialization, to ensure unit variance between each block.

            beta: <float> - The beta value, used to scale tensor variances for initialization, 
            to ensure unit variance between each block.

            stoch_depth: <float> - The stochastic depth skip probability for the residual path
            in this stage.

            is_transition: <bool> - Force this block to be configured as a NFNet transition block.

        Return:
            list(list(tf.layers.Layer)) - A four element list comprised of:
                [
                    input_layers, <= Layers to be applied directly to the input
                    residual_layers, <= Layers applied after input layers, forming the residual path
                    skip_layers,  <= Layers applied after input layers, forming the skip path
                    output_layers <= Layers to aggregate the residual and skip path, to apply prior to the output
                ]
        """

        # Any difference in dimensions (either across frequency, or across the channel dimension),
        # means that this is an NFNet transition block. There is also the option to force this block
        # to be configured as an NFNet transition block, with matching i/o dimension, which is required
        # in the first nfnet block on the slow path in [1], which has matching i/o channels (due to 
        # concatenation with the fast path), and no frequency downsampling.
        is_transition_block = (freq_downsample > 1) or (input_channels != output_channels) or is_transition

        #
        # Input layers + initialize the residual path depending on whether this is a transition block.
        #
        if is_transition_block:
            input_layers = [
                tf.keras.layers.Activation(_scaled_activation(self._scaled_activation_type)),
                ScalarMultiply(beta)
            ]
            residual_path = []
        else:
            input_layers = []
            residual_path = [
                tf.keras.layers.Activation(_scaled_activation(self._scaled_activation_type)),
                ScalarMultiply(beta)
            ]

        #
        # Path 1 (residual)
        # 
        strides = [[1,1], [freq_downsample,1], [1,1], [1,1]] # NOTE [matt.c.mccallum 03.11.22]: Should this frequency downsample be on the frequency convolution layer, or the first (time convoluation) layer? The original NFNet paper talks about striding on the first 3 x 3 conv layer, but in this case, freq convs don't happen until the second non-1x1 conv layer.
        per_layer_out_chans = [output_channels//2]*3 + [output_channels]
        groups = [1] + [output_channels//2//group_size]*2 + [1]
        activations = [_scaled_activation(self._scaled_activation_type)]*(len(kernels)-1) + [None]
        residual_path += list(itertools.chain(*[
            self.get_conv(
                channels=c,
                kernel=k,
                strides=s,
                groups=g,
                activation=a
            ) for c, k, s, g, a in zip(per_layer_out_chans, kernels, strides, groups, activations)
        ])) + [
            self._make_squeeze_and_excite(per_layer_out_chans[-1]),
            ScalarMultiply(0.0, learnable=True),
            ScalarMultiply(alpha)
        ]

        #
        # Path 2 (skip)
        # 
        skip_path = []
        if freq_downsample > 1: # <= Only include average pooling when we are downsampling
            skip_path += [
                tf.keras.layers.AveragePooling2D(
                    pool_size=[freq_downsample, 1], # Should this be 1 or 2 on the time dimension? Seeing we only subsample the frequency dimension...
                    strides=[freq_downsample, 1],
                    padding='same'
                )
            ]
        if is_transition_block:
            skip_path += [
                WeightStandardization(tf.keras.layers.Conv2D(
                    filters=output_channels,
                    kernel_size=[1, 1],
                    strides=[1, 1],
                    padding='same',
                    groups=1,
                    activation='linear', # NOTE [matt.c.mccallum 03.11.22]: No activation on shortcut path
                    kernel_initializer=tf.keras.initializers.VarianceScaling(
                        scale=1.0,
                        mode='fan_in',
                        distribution='untruncated_normal'
                    )
                ))
            ]

        #
        # Concat / Add to provide the output
        #
        output_layers = [
            StochDepth(survival_probability=1-stoch_depth, scale_during_test=False)
        ]

        return [input_layers, residual_path, skip_path, output_layers]
    
    def _apply_nfnet_block(self, nfnet_block, input):
        """
        Apply a single NFNet block to the provided input tensor.

        Args:
            nfnet_block: <list(list(tf.keras.layers.Layer))> - Layers constructed by the 
            `_make_nfnet_block` method. This should include four lists of layers, comprising
            the input layers, residual layers, skip layers, and output layers.

            input: tf.Tensor - The input tensor to apply the layers to.

        Return:
            <tf.Tensor> - The output, after applying the stem module layers.
        """
        # Input path
        for lyr in nfnet_block[0]:
            input = lyr(input)

        # Residual path
        conv = input
        for idx, lyr in enumerate(nfnet_block[1]):
            # NOTE [matt.c.mccallum 03.18.22]: Squeeze and excite layer is always third 
            # to last on the conv path, and requires custom application. Once we move these
            # modules to their own layer class, with a call method, this can be removed.
            if len(nfnet_block[1])-idx == 3:
                conv = self._apply_squeeze_and_excite(lyr, conv)
            else:
                conv = lyr(conv)

        # Skip path
        skip = input
        for lyr in nfnet_block[2]:
            skip = lyr(skip)

        # Output
        output = nfnet_block[3][0]([skip, conv])
        return output

    def _make_squeeze_and_excite(self, output_channels):
        """
        Create a squeeze and excite module [1].

        Args:
            output_channels: <int> - The number of channels at the output of the squeeze and excitation block.

        Return:
            <list(tf.keras.layers.Layer)> - The layers comprising the squeeze and excite module.

        References:
            [1] Hu, Jie, Li Shen, and Gang Sun. "Squeeze-and-excitation networks." In Proceedings of the 
            IEEE conference on computer vision and pattern recognition, pp. 7132-7141. 2018.
        """
        # NOTE [matt.c.mccallum 03.22.22]: To mirror closely the NFNet reference code produced by DeepMind,
        #      we apply a regular relu and sigmoid here as activation, rather than the scaled activation
        #      employed throughout the rest of the NFNet block.
        return [
            tf.keras.layers.GlobalAveragePooling2D(keepdims=False),
            tf.keras.layers.Dense(units=output_channels//2, activation='relu', use_bias=True), # Squeeze excite layers usually have half the number of hidden units to output units, but we should make this configurable
            tf.keras.layers.Dense(units=output_channels, activation='sigmoid', use_bias=True)
            # NOTE [matt.c.mccallum 03.12.22]: Broadcasting of the output is done when this layer is applied
        ]

    def _apply_squeeze_and_excite(self, se_lyrs, input):
        """
        Apply a squeeze and excite module to an input.

        Args:
            se_lyrs: <list(tf.keras.layers.Layer)> - The layers comprising the squeeze and 
            excite module

            input: <tf.Tensor> - The input tensor to apply the layers to.

        Return:
            <tf.Tensor> - The output after the module has been applied to the input.
        """
        output = input

        for lyr in se_lyrs:
            output = lyr(output)

        # NOTE [matt.c.mccallum 03.12.22]: Re-insert time and frequency dimensions below.
        output = tf.expand_dims(tf.expand_dims(output*2.0, axis=1), axis=1)*input
        return output

    def _make_fast_to_slow_fusion(self, time_kernel_length, time_stride, input_channels, output_channels):
        """
        Make layers that comprise the operations in order to fuse the fast path of NFNet stages to the
        slow path. This simply applies a Tx1 time-strided convolution to the fast path, to match the 
        dimensionality of the slow path, then concatenates with the slow path.

        Args:
            time_kernel_length: <int> - The length of the convolutional kernel in time.

            time_stride: <int> - The size of the stride along the time dimension.

            output_channels: <int> - The number of channels of the fast path at the output
            of the convolutional layer, prior to concatenation with the slow path.

        Return:
            list(tf.keras.layers.Layer) - The layers in the fast to slow fusion module.
        """

        return [
            WeightStandardization(tf.keras.layers.Conv2D(
                filters=input_channels,
                kernel_size=[1, time_kernel_length],
                strides=[1, time_stride],
                padding='same',
                activation='linear', # NOTE [matt.c.mccallum 03.11.22]: No activation on shortcut path
                kernel_initializer=tf.keras.initializers.VarianceScaling(
                    scale=1.0,
                    mode='fan_in',
                    distribution='untruncated_normal'
                )
            )),
            WeightStandardization(tf.keras.layers.Conv2D(
                filters=output_channels,
                kernel_size=[1, 1],
                strides=[1, 1],
                padding='same',
                activation='linear', # NOTE [matt.c.mccallum 03.11.22]: No activation on shortcut path
                kernel_initializer=tf.keras.initializers.VarianceScaling(
                    scale=1.0,
                    mode='fan_in',
                    distribution='untruncated_normal'
                )
            )),
            tf.keras.layers.Concatenate(axis=-1), # Concatenates on the last (channel) dimension by default
        ]

    def _apply_fast_to_slow_fusion(self, fusion_module, slow, fast):
        """
        Apply the fast-to-slow fusion module to create the next (fused) input for
        the slow path.

        Args:
            fusion_module: <list(tf.keras.layers.Layer)> - The layers comprising the fusion module.

            slow: <tf.Tensor> - The output of the previous stage form the slow path, to be fused
            with the fast path.

            fast: <tf.Tensor> - The output of the previous stage of the fast path, to be fused into
            the slow path.

        Return:
            <tf.Tensor> - The output of the fast-to-slow fusion, to be used as input to the next
            stage in the slow path.
        """
        fast = fusion_module[0](fast)
        fast = fusion_module[1](fast)
        return fusion_module[2]([slow, fast])



# coding=utf-8
# Copyright 2021 Pandora Media, LLC.
#
# Licensed under the GNU GPL License, Version 3.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gnu.org/licenses/gpl-3.0.en.html
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python standard library imports
# None

# Third party imports
import tensorflow as tf
from scooch import Param

# Local imports
from .lr_schedule import LRSchedule


class WarmupSchedule(LRSchedule, tf.keras.optimizers.schedules.LearningRateSchedule):
    """
    Adds a linear warmup to a learning rate schedule.
    """

    _cold_learning_rate = Param(
        float,
        default=0.0,
        doc="Learning rate to start the warmup with"
    )
    _warm_learning_rate = Param(
        float,
        doc="Desired learning rate after the warmup"
    )
    _warmup_steps = Param(
        int,
        doc="Number of warmup steps"
    )
    _schedule = Param(
        LRSchedule,
        doc="Original learning rate schedule"
    )
    _initial_step = Param(
        int,
        default=0,
        doc="The step to start the schedule out - higher integers will start training at steps later in the schedule."
    )

    def init(self):
        self._warmup = tf.keras.optimizers.schedules.PolynomialDecay(
            initial_learning_rate=self._cold_learning_rate,
            decay_steps=self._warmup_steps,
            end_learning_rate=self._warm_learning_rate,
        )
        self._schedule.init()

    def __call__(self, step):
        """
        Get the learning rate for a step.

        Args:
            step: TensorFlow Tensor - the current training step.
        """
        # NOTE [matt.c.mccallum 04.26.22]: Keras's model.fit() always passes 0 on the first step to the learning
        #      rate schedule, regardless of the `initial_epoch` argument. We add an offset here so that we can
        #      customize where the learning rate schedule begins.
        step = step + self._initial_step
        return tf.cond(
            step < self._warmup_steps,
            lambda: self._warmup(step),
            lambda: self._schedule(step - self._warmup_steps)
        )

    def get_config(self):
        return self.cfg

    @classmethod
    def from_config(cls, cfg):
        x = cls(cfg)
        x.init()
        return x
        


# coding=utf-8
# Copyright 2021 Pandora Media, LLC.
#
# Licensed under the GNU GPL License, Version 3.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gnu.org/licenses/gpl-3.0.en.html
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Local imports
# None.

# Third party imports
from scooch import Configurable

# Python standard library imports
# None.


class TrainMetric(Configurable):
    """
    A base class for keras or custom metrics.
    """

    def init(self):
        """
        Compile the metric function. Separated from constructor so that it may be called
        outside of the SCOOCH class heirarchy instantiation.
        """
        self.initialize_base()




# coding=utf-8
# Copyright 2021 Pandora Media, LLC.
#
# Licensed under the GNU GPL License, Version 3.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gnu.org/licenses/gpl-3.0.en.html
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python standard library imports
import urllib.parse as urlparse

# Third party imports
import tensorflow as tf
from scipy.special import logsumexp
import numpy as np
from scooch import Param

# Local imports
from .transform_feature import TransformFeature


class EmbeddingFeature(TransformFeature):
    """
    A feature that applies an ML model to input data to create an embedding.
    """

    #
    # SCOOCH Configuration
    #
    _model_location = Param(
        str,
        doc="Location of the model to load and analyze feature data with."
    )
    _apply_softmax = Param(
        bool,
        default=False,
        doc="Whether to apply a softmax to the model output feature."
    )

    _model = None

    _CHUNK_SIZE = 8000

    # 
    # Methods
    #
    def __init__(self, cfg):
        """
        *Constructor*

        Args:
            cfg: scooch.Config - The scooch configuration for this object.
        """
        super().__init__(cfg)
        self.load_model()

    def _extract(self, source_feature, start_time, chunk_size):
        """
        Extracts feature data by applying a model to samples of an input feature, over
        a given time period.

        Args:
            source_feature: mule.features.Feature - The feature to transform.

            start_time: int - The index in the feature at which to start extracting / transforming.

            chunk_size: int - The length of the chunk following the `start_time` to extract
            the feature from.
        """
        data = super()._extract(source_feature, start_time, chunk_size)

        if len(data):
            data = self._model.predict(
                    [data],
                    callbacks=None
                )
            data = data.T
            if self._apply_softmax:
                data = np.nan_to_num(np.exp(data - logsumexp(data, axis=0)))
            return data
        else:
            return None

    def load_model(self):
        """
        Loads the model that this feature uses from a configured location.
        """
        # If GS use google-cloud-storage...
        parsed_loc = urlparse.urlparse(self._model_location)
        if parsed_loc.scheme == 'gs':
            # TODO [matt.c.mccallum 11.21.22]: Write google storage download code here.
            pass
        # Otherwise assume local
        self._model = tf.keras.models.load_model(self._model_location, compile=False)



# coding=utf-8
# Copyright 2024 Pandora Media, LLC.
#
# Licensed under the GNU GPL License, Version 3.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gnu.org/licenses/gpl-3.0.en.html
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python standard library imports
# None.

# Third party imports
from scooch import Param
import numpy as np

# Local imports
from .extractor import Extractor


class SliceExtractor(Extractor):
    """
    An extractor that extracts slices of feature data each surrounding an
    index in the provided input feature.
    """

    #
    # SCOOCH Configuration
    #
    _hop = Param(
        int,
        default=150,
        doc="How often to extract a slice of data from the input when extracting over a time range, in indices."
    )
    _look_forward = Param(
        int,
        default=150,
        doc="How many indices ahead of the index to be extracted from, to include in a given slice."
    )
    _look_backward = Param(
        int,
        default=150,
        doc="How many indices behind the index to be extract from, to include in a given slice"
    )
    _standard_normalize = Param(
        bool,
        default=True,
        doc="Whether to standard normalize each slice of data that extracted from the input feature."
    )

    # 
    # Methods
    #
    def extract_range(self, feature, start_index, end_index):
        """
        Extracts data over a given index range from a single feature. This will
        extract regularly spaced 2D slices of data from the input feature. Note
        that first feature will be extracted at an integer multiple of the extractor's
        configured `hop` parameter from the beginning of the feature.

        Args:
            feature: mule.feature.Feature - A feature to extract data from.

            start_index: int - The first index (inclusive) at which to start extracting
            slices.

            end_index: int - The last index (exclusive) at which to return data.

        Return:
            numpy.ndarray - The extracted feature data. Time on the first axis, features
            on the remaining axes.
        """
        indices = [time for time in range(0, end_index, self._hop) if time > start_index]
        features = [feature]*len(indices)
        if len(features)==0:
            return np.empty((0,0))
        return self.extract_batch(features, indices)
        
    def extract_batch(self, features, indices):
        """
        Extracts a batch of slices of data from a range of features, centered on specific 
        indices.

        Args:
            features: list(mule.features.Feature) - A list of features from which to extract 
            data from.

            indices: list(int) - A list of indices, the same size as `features`. Each element
            provides an index at which to extract data from the coressponding element in the
            `features` argument.

        Return:
            np.ndarray - A batch of features, with features on the batch dimension on the first
            axis and feature data on the remaining axes.
        """
        indices = [idx if idx >= self._look_backward else self._look_backward for idx in indices]
        indices = [idx if idx <= len(feat) - self._look_forward else len(feat) - self._look_forward for idx, feat in zip(indices, features)]
        samples = [feature.data[:, (idx-self._look_backward):(idx+self._look_forward)] for idx, feature in zip(indices, features)]
        samples = [x.reshape((1, *x.shape, 1)) for x in samples] # Add batch and channel dimensions
        samples = np.vstack(samples)

        if self._standard_normalize:
            samples -= np.mean(samples, axis=(1,2,3), keepdims=True)
            all_vars = np.std(samples, axis=(1,2,3), keepdims=True)
            all_vars = np.nan_to_num(all_vars, nan=1.0, posinf=1.0, neginf=1.0)
            all_vars = np.maximum(all_vars, 0.01)
            samples /= all_vars 

        return samples



# coding=utf-8
# Copyright 2021 Pandora Media, LLC.
#
# Licensed under the GNU GPL License, Version 3.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gnu.org/licenses/gpl-3.0.en.html
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python standard library imports
# None.

# Third party imports
import numpy as np
import librosa
from scooch import Param

# Local imports
from .source_feature import SourceFeature


class AudioWaveform(SourceFeature):
    """
    A feature representing an audio waveform, extracted from an audio file.
    """

    #
    # SCOOCH Configuration
    #
    _sample_rate = Param(
        int,
        default=16000,
        doc="The sample rate of the output waveform. Input audio files will be resampled to this."
    )
    # TODO [matt.c.mccallum 11.18.22]: SCOOCH does not support virtual params at the moment. For now
    #      it is up to the user to ensure the input type is an AudioFile, but once virtual parameter
    #      support is added to SCOOCH, we should override the `_input_file` from the SourceFeature
    #      base class parameter here, to ensure that is of an AudioFile type.

    #
    # Methods
    #
    def _extract(self, source, start_time, chunk_size):
        """
        Extracts data from the audio waveform from a source file, for a given number of samples.

        Args:
            source: mule.source_files.AudioFile - An audio file to extract data from.

            start_time: int - The sample index at which to start extracting an audio chunk from.
            Currently this is unused, and audio extraction will continue from where it last left
            off.

            chunk_size: int - The number of samples to extract in this call.

        Return:
            np.ndarray - The extracted audio samples, summed to mono. Sample index is on the second
            axis.
        """
        # TODO [matt.c.mccallum 11.19.22]: Make this function work for non-sequential reads, i.e., respect the start_time argument.
        # Read from file
        data = source.read(chunk_size)
        # Sum to mono
        samples = np.sum(data, axis=0)
        # Resample
        if source.sample_rate != self._sample_rate:
            samples = librosa.core.resample(samples, source.sample_rate, self._sample_rate)
        # TODO [matt.c.mccallum 11.18.22]: Improve handling at chunk boundaries
        return samples.reshape((1, -1))

    # 
    # Properties
    #
    @property
    def sample_rate(self):
        """
        The sample rate of the waveform feature itself. (Any audio file will be resampled
        to this).
        """
        return self._sample_rate



# coding=utf-8
# Copyright 2021 Pandora Media, LLC.
#
# Licensed under the GNU GPL License, Version 3.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gnu.org/licenses/gpl-3.0.en.html
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python standard library imports
# None.

# Third party imports
import tensorflow as tf
from scooch import Param

# Local imports
from .loss import Loss


class SimClrLoss(Loss, tf.keras.losses.Loss):
    """
    Implements the loss function described in [1]. Intended for unsupervised learning via
    augmented data pairs.

    [1] - Ting, C., et. al "A Simple Framework for Contrastive Learning of Visual Representations",
        arXiv preprint arXiv:2002.05709, 2020.

    """

    #
    # SCOOCH Configuration
    #
    _temperature = Param(
        float,
        default=0.1,
        doc="The SimCLR loss temperature. (See [1] in class docstring)."
    )

    _LARGE_NUMBER = 1e10

    # 
    # Methods
    #
    def __init__(self, cfg, *args, **kwargs):
        """
        Constructor.
 
        Args:
            cfg: Config - The SCOOCH loss's configuration.
        """
        super().__init__(cfg, *args, **kwargs)

        # NOTE [matt.c.mccallum 02.10.22]: Remove any keras based reduction / normalization across GPUs, we handle this ourselves
        self._star_kwargs.update({
            'reduction': tf.keras.losses.Reduction.SUM
        })

    @staticmethod
    def _cross_replica_concat(tensor, strategy=None):
        """Reduce a concatenation of the `tensor` across devices.
        Args:
            tensor: tensor to concatenate.
            strategy: A `tf.distribute.Strategy`. If not set, CPU execution is assumed.
        Returns:
            Tensor of the same rank as `tensor` with first dimension `num_replicas`
            times larger.

        Taken from:
            https://github.com/google-research/simclr/blob/master/tf2/objective.py
        """
        if strategy is None or strategy.num_replicas_in_sync <= 1:
            return tensor

        num_replicas = strategy.num_replicas_in_sync

        replica_context = tf.distribute.get_replica_context()
        with tf.name_scope('cross_replica_concat'):
            # This creates a tensor that is like the input tensor but has an added
            # replica dimension as the outermost dimension. On each replica it will
            # contain the local values and zeros for all other values that need to be
            # fetched from other replicas.
            ext_tensor = tf.scatter_nd(
                indices=[[replica_context.replica_id_in_sync_group]],
                updates=[tensor],
                shape=tf.concat([[num_replicas], tf.shape(tensor)], axis=0)
            )

            # As every value is only present on one replica and 0 in all others, adding
            # them all together will result in the full tensor on all replicas.
            ext_tensor = replica_context.all_reduce(tf.distribute.ReduceOp.SUM, ext_tensor)

            # Flatten the replica dimension.
            # The first dimension size will be: tensor.shape[0] * num_replicas
            # Using [-1] trick to support also scalar input.
            return tf.reshape(ext_tensor, [-1] + ext_tensor.shape.as_list()[2:])

    def call(self, y_true, y_pred):
        """
        Compute the loss function.

        Args:
            y_true: <tf.Tensor> - Unused. This is an unsupervised SimCLR loss and so uses no target data.

            y_pred: list(<tf.Tensor>) - Two model outputs in the order: anchor_output, augmented_output.
        """
        # Collect examples from all GPUs
        strat = tf.distribute.get_strategy()

        # Assume the output layer of the network is already l2 normalized... No need for the line below 
        y_pred = tf.keras.backend.l2_normalize(y_pred, axis=-2)
        per_device_batch_size = tf.shape(y_pred)[0]
        
        all_examples = self._cross_replica_concat(y_pred, strat)
        raw, augmented = all_examples[:,:,0], all_examples[:,:,1]
        total_batch_size = tf.shape(raw)[0]
        collected_examples = tf.concat([raw, augmented], axis=0)
        collected_examples = tf.squeeze(collected_examples)

        # Compute logits
        all_sims = tf.tensordot(collected_examples, tf.transpose(collected_examples), axes=1)/self._temperature
        all_sims -= tf.linalg.diag(tf.ones(tf.shape(all_sims)[0]))*self._LARGE_NUMBER
        labels = tf.linalg.diag(tf.ones([total_batch_size]))
        empty = tf.zeros([total_batch_size, total_batch_size])
        labels = tf.concat([tf.concat([empty, labels], axis=1), tf.concat([labels, empty], axis=1)], axis=0)

        # Get index of device
        replica_context = tf.distribute.get_replica_context()
        replica_id = tf.cast(
            tf.cast(replica_context.replica_id_in_sync_group, tf.uint32), tf.int32)
        start_idx = replica_id * per_device_batch_size
        end_idx = (replica_id+1) * per_device_batch_size

        # Just get the losses that are relevant to the current device
        # TODO [matt.c.mccallum 01.23.22]: More efficient to do this before computing the logits above,
        #      rather than computing all logits on every device. But it's a drop in the ocean compared
        #      to the networks we're training.
        this_all_a_sims = all_sims[start_idx : end_idx]
        this_a_labels = labels[start_idx : end_idx]
        this_all_b_sims = all_sims[start_idx + total_batch_size : end_idx + total_batch_size]
        this_b_labels = labels[start_idx + total_batch_size : end_idx + total_batch_size]

        loss_a = tf.nn.softmax_cross_entropy_with_logits(this_a_labels, this_all_a_sims)
        loss_b = tf.nn.softmax_cross_entropy_with_logits(this_b_labels, this_all_b_sims)
        loss = tf.reduce_mean(loss_a + loss_b)

        num_replicas = strat.num_replicas_in_sync
        loss = loss/num_replicas

        return loss




# coding=utf-8
# Copyright 2024 Pandora Media, LLC.
#
# Licensed under the GNU GPL License, Version 3.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gnu.org/licenses/gpl-3.0.en.html
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python standard library imports
# None.

# Third party imports
from scooch import Param
from scooch import Configurable
from scooch import ConfigList

# Local imports
from .features import SourceFeature
from .features import TransformFeature


class Analysis(Configurable):
    """
    A class encapsulating analysis of a single input file.
    """

    #
    # SCOOCH Configuration
    #
    _source_feature = Param(
        SourceFeature,
        doc="The feature used to decode the provided raw file data."
    )
    _feature_transforms = Param(
        ConfigList(TransformFeature),
        doc="Feature transformations to apply, in order, to the source feature generated from the input file."
    )

    #
    # Methods
    #
    def analyze(self, fname):
        """
        Analyze features for a single filepath.

        Args:
            fname: str - The filename path, from which to generate features.

        Return:
            mule.features.Feature - The feature resulting from the configured feature
            transformations.
        """
        for feat in [self._source_feature] + self._feature_transforms:
            feat.clear()

        self._source_feature.from_file(fname)
        input_feature = self._source_feature
        for feature in self._feature_transforms:
            feature.from_feature(input_feature)
            input_feature = feature

        return input_feature



# coding=utf-8
# Copyright 2021 Pandora Media, LLC.
#
# Licensed under the GNU GPL License, Version 3.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gnu.org/licenses/gpl-3.0.en.html
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python standard library imports
# None.

# Third party imports
import numpy as np

# Local imports
from .transform_feature import TransformFeature


class TimelineAverage(TransformFeature):
    """
    A feature that averages over data segments extracted from the input feature.
    """

    _feature_count = 0

    #
    # Methods
    #
    def _extract(self, source_feature, start_time, chunk_size):
        """
        Extracts an average of feature segments from the input feature over the provided time range.

        Args:
            source_feature: mule.features.AudioWaveform - The feature to transform into a mel spectrogram

            start_time: int - The index in the feature at which to start extracting / transforming.

            chunk_size: int - The length of the chunk following the `start_time` to extract
            the feature from.
        """
        data = np.squeeze(super()._extract(source_feature, start_time, chunk_size))

        self._feature_count += data.shape[1]

        # TODO [matt.c.mccallum 12.15.22]: Handle overflow here for large features
        data = np.sum(data, axis=1, keepdims=True)

        return data

    def add_data(self, data):
        """
        Adds data to previously accumulated data via summation.

        Arg:
            data: np.ndarray - The data to be appended to the object's summation.
        """
        if not hasattr(self, '_data') or self._data is None:
            self._data = np.zeros(data.shape)

        self._data = self._data + data
        
    @property
    def data(self):
        """
        The average of feature segments seen so far.
        """
        return self._data/self._feature_count



# coding=utf-8
# Copyright 2024 Pandora Media, LLC.
#
# Licensed under the GNU GPL License, Version 3.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gnu.org/licenses/gpl-3.0.en.html
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python standard library imports
# None.

# Third party imports
import click

# Local imports
# None.


def partial_option(*args, **kwargs):
    option = click.option(*args, **kwargs)

    def option_decorator(command=None):
        if command:
            # We were used as a decorator without arguments, and now we're being
            # called on the command function.
            return option(command)
        else:
            # We are being called with empty parens to construct a decorator.
            return option

    return option_decorator

#
# Options below
#
config = partial_option(
        "--config", 
        "-c",
        help="Path to yaml scooch file defining an analysis pipeline"
    )

input_file = partial_option(
        "--input-file", 
        "-i",
        help="Path to input audio file to be analyzed"
)

output_file = partial_option(
        "--output-file", 
        "-o",
        help="Path to output file location to place resulting analyzed features"
    )



# coding=utf-8
# Copyright 2021 Pandora Media, LLC.
#
# Licensed under the GNU GPL License, Version 3.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gnu.org/licenses/gpl-3.0.en.html
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python standard library imports
import inspect
import sys

# Third party imports
import tensorflow as tf
from scooch import configurize

# Local imports
from .loss import Loss
from .sim_clr_loss import SimClrLoss


# Augment all Tensorflow loss functions with the Configurable class
clsmembers = inspect.getmembers(sys.modules[tf.keras.losses.__name__], inspect.isclass)
for mem in clsmembers:
    if mem[0] != 'Loss':
        tf_class = configurize(mem[1], Loss, init_base_on_construction=False)
        exec(mem[0] + ' = tf_class')
        


# coding=utf-8
# Copyright 2021 Pandora Media, LLC.
#
# Licensed under the GNU GPL License, Version 3.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gnu.org/licenses/gpl-3.0.en.html
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python standard library imports
# None.

# Third party imports
from scooch import Param

# Local imports
from ...extractors import Extractor
from ..feature import Feature


class TransformFeature(Feature):
    """
    Base class for all features that are transforms of other features.
    """

    #
    # SCOOCH Configuration
    #
    _extractor = Param(
        Extractor,
        doc="An object defining how data will be extracted from the input feature and provided to the transformation of this feature."
    )

    # The size in time of each chunk that this feature will process at any one time.
    _CHUNK_SIZE = 44100*60*15

    #
    # Methods
    #
    def from_feature(self, source_feature):
        """
        Populates this features data as a transform of the provided input feature.

        Args:
            source_feature: mule.features.Feature - A feature from which this feature will
            be created as a transformation thereof.
        """
        boundaries = list(range(0, len(source_feature), self._CHUNK_SIZE)) + [len(source_feature)]
        chunks = [(start, end) for start, end in zip(boundaries[:-1], boundaries[1:])]
        for start_time, end_time in chunks:
            data = self._extract(source_feature, start_time, end_time-start_time)
            if data is not None and len(data):
                self.add_data(data)

    def _extract(self, source_feature, start_time, chunk_size):
        """
        Extracts feature data as a transformation of a given source feature for a given
        time-chunk.

        Args:
            source_feature: mule.features.Feature - The feature to transform.

            start_time: int - The index in the feature at which to start extracting / transforming.

            chunk_size: int - The length of the chunk following the `start_time` to extract
            the feature from.
        """
        end_time = start_time + chunk_size
        return self._extractor.extract_range(source_feature, start_time, end_time)

    def __len__(self):
        if hasattr(self, '_data'):
            return self._data.shape[1]
        else:
            return 0
        


# coding=utf-8
# Copyright 2021 Pandora Media, LLC.
#
# Licensed under the GNU GPL License, Version 3.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gnu.org/licenses/gpl-3.0.en.html
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python standard library imports
# None.

# Third party imports
import tensorflow as tf
import warnings

# Local imports
# None.


class WeightStandardization(tf.keras.layers.Wrapper):
    """
    Applies weight scaled normalization [1] to a convolutional or dense layer, which is very similar
    to centered (zero-mean) weight-norm layers, however, has no learnable magnitude, forcing unit
    variance and zero mean in the weights.

    This code is largely inspied by the WeightNorm layer found in the tensorflow_addons package:

        https://www.tensorflow.org/addons/tutorials/layers_weightnormalization

    and that in the tensorflow probability package:

        https://www.tensorflow.org/probability/api_docs/python/tfp/layers/weight_norm/WeightNorm

    Example:

        ```python
        net = WeightStandardization(tf.keras.layers.Conv2D(2, 2, activation='relu'), input_shape=(32, 32, 3))(x)
        net = WeightStandardization(tf.keras.layers.Dense(120, activation='relu'))(net)
        net = WeightStandardization(tf.keras.layers.Dense(num_classes))(net)
        ```

    References:
    
        [1] Qiao, Siyuan, Huiyu Wang, Chenxi Liu, Wei Shen, and Alan Yuille. "Micro-batch training with batch-channel 
        normalization and weight standardization." arXiv preprint arXiv:1903.10520 (2019).
    """

    def __init__(self, layer, **kwargs):
        """
        **Constructor**
        
        Args:

          layer: <tf.keras.layers.Layer> - Supported layer types are `Dense`, `Conv2D`, and 
          `Conv2DTranspose`. Layers with multiple inputs are not supported.

          **kwargs: <dict(str, object)> Additional keyword args passed to `tf.keras.layers.Wrapper`.
        """
        if not isinstance(layer, tf.keras.layers.Layer):
            raise ValueError(
                'Please initialize `WeightNorm` layer with a `tf.keras.layers.Layer` '
                'instance. You passed: {input}'.format(input=layer))

        layer_type = type(layer).__name__
        # TODO [matt.c.mccallum 03.12.22]: Support Conv2DTranspose layers, by supporting channels
        #      on the first dimension.
        if layer_type not in ['Dense', 'Conv2D']:
            warnings.warn('`WeightNorm` is tested only for `Dense`, and `Conv2D` layers.'
                          'You passed a layer of type `{}`'
                          .format(layer_type))

        super(WeightStandardization, self).__init__(layer, **kwargs)

        self.filter_axis = -1

        self._track_trackable(layer, name='layer')

    def _compute_weights(self):
        """
        Generate weights with standardization.
        """
        
        weight = self.v

        mean = tf.math.reduce_mean(weight, axis=(0, 1, 2), keepdims=True)
        var = tf.math.reduce_variance(weight, axis=(0, 1, 2), keepdims=True)
        fan_in = tf.cast(tf.math.reduce_prod(weight.get_shape()[:-1]), tf.float32) # Batch dimension?

        # Manually fused normalization, eq. to (w - mean) * gain / sqrt(N * var)
        eps = tf.constant(0.0001)
        scale = tf.math.rsqrt(tf.math.maximum(var * fan_in, eps)) * self.gain
        shift = mean * scale
        self.layer.kernel = weight * scale - shift

    def build(self, input_shape=None):
        """
        Build `Layer`.

        Args:

          input_shape: The shape of the input to `self.layer`.

        Raises:

          ValueError: If `Layer` does not contain a `kernel` of weights
        """

        input_shape = tf.TensorShape(input_shape).as_list()
        input_shape[0] = None
        self.input_spec = tf.keras.layers.InputSpec(shape=input_shape)

        if not self.layer.built:
            self.layer.build(input_shape)

            if not hasattr(self.layer, 'kernel'):
                raise ValueError('`WeightStandardization` must wrap a layer that'
                                ' contains a `kernel` for weights')

            kernel_norm_axes = list(range(self.layer.kernel.shape.rank))
            kernel_norm_axes.pop(self.filter_axis)
            # Convert `kernel_norm_axes` from a list to a constant Tensor to allow
            # TF checkpoint saving.
            self.kernel_norm_axes = tf.constant(kernel_norm_axes)

            # to avoid a duplicate `kernel` variable after `build` is called
            self.v = self.layer.kernel
            self.layer.kernel = None

            # Get gain
            # NOTE [matt.c.mccallum 03.12.22]: No need to broadcast this, it should
            #      match the var dimensions
            self.gain = tf.Variable(initial_value=tf.ones(self.v.get_shape()[-1], dtype=self.v.dtype), trainable=True)

        super(WeightStandardization, self).build()

    def call(self, inputs):
        """
        Call `Layer`.
        
        Args:
            inputs: <tf.Tensor> - The input tensor to apply the weight standardized layer to.

        Return:
            <tf.Tensor> - The tensor at the output of the layer.
        """
        self._compute_weights()
        output = self.layer(inputs)
        return output

    def compute_output_shape(self, input_shape):
        """
        Computes the output shape after the application of `Layer`.

        Args:
            input_shape: <tf.Tensor> - The shape of the input to this layer.

        Return:
            <tf.Tensor> - The shape after the application of this layer.
        """
        return tf.TensorShape(
            self.layer.compute_output_shape(input_shape).as_list())



# coding=utf-8
# Copyright 2021 Pandora Media, LLC.
#
# Licensed under the GNU GPL License, Version 3.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gnu.org/licenses/gpl-3.0.en.html
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python standard library imports
import tempfile
import io
import subprocess
import struct
import wave
import os

# Third party imports
import numpy as np
from scooch import Param

# Local imports
from .source_file import SourceFile


class AudioFile(SourceFile):
    """
    A class for reading audio files.
    """

    _sample_rate = Param(
        int,
        default=44100,
        doc="The sample rate that an input audio file is interpreted as"
    )
    # TODO [matt.c.mccallum 11.18.22]: Handle sampling rates here

    _wav_file = None
    _use_local_file = False
    _max_len = 60*30 # 30 minutes
    _bit_depth = 16
    _samp_rate = 44100
    _wav_ffmpeg_fmt = 'pcm_s' + str(_bit_depth) + 'le'

    # Map file extensions to ffmpeg formats here
    _FORMATS = {
        '.mp3': 'mp3',
        '.wav': 'wav',
        '.wave': 'wav'
    }

    def __del__(self):
        """
        Destructor.
        """
        if self._wav_file is not None:
            self._wav_file.close()
        
    def load(self, file):
        """
        Converts the file to a temporary wav file. Once converted this wav file will stick around as long as this object
        exists.
        Note: Depending on whether use_local_file is true, this will either read directly from the system pipes for output
        (false), or save the result to a temporary file on disk, transcode, then read the temporary file (true).

        Args:
            file: file - A file object containing mp3 data to decode.
        """
        ext = os.path.splitext(file)[1]
        if ext in self._FORMATS:
            format = self._FORMATS[ext]
        else:
            raise IOError(f"Unsupported audio file extension: {ext}, supplied to audio file reader")

        if self._use_local_file:
            self._tempf = tempfile.NamedTemporaryFile(mode='r+b', suffix='.wav')
            cmd = f"ffmpeg -y -loglevel panic -t {self._max_len} -f {format} -i {file} -map_metadata -1 " \
                f"-vn -acodec {self._wav_ffmpeg_fmt} -ar {self._samp_rate} -f wav {self._tempf.name}"
            subprocess.run(cmd.split(" "))

                # Create a wav object containing the decoded data.
            self._tempf.seek(0)
            self.wav_file = self._tempf
        else:
            with open(file, 'rb') as f:
                cmd = f"ffmpeg -loglevel panic -t {str(self._max_len)} -f {format} -i pipe:0 -map_metadata -1 -vn " \
                    f"-acodec {self._wav_ffmpeg_fmt} -ar {self._samp_rate} -f wav pipe:1"
                x = subprocess.run(cmd.split(" "), input=f.read(), capture_output=True)
                in_mem_file = io.BytesIO(x.stdout)

            # Fix file size as ffmpeg output via std stream doesn't include a file size.
            in_mem_file.seek(0)
            file_length = in_mem_file.seek(0, 2)
            in_mem_file.seek(4)
            in_mem_file.write(struct.pack('i', file_length - 8))
            in_mem_file.seek(0)
            test_data = in_mem_file.read(10000)
            data_start = test_data.find(b'data')
            in_mem_file.seek(data_start + 4)
            in_mem_file.write(struct.pack('i', file_length - data_start - 8))
            in_mem_file.seek(0)

            self.wav_file = in_mem_file

    def close(self):
        """
        Closes any previously loaded file.
        """
        if self._wav_file is not None:
            self._wav_file.close()
        if hasattr(self, '_tempf'):
            del self._tempf

    def _packing_string( self, num_frames ):
        """
        Get the string for packing or unpacking a given number of frames using the struct module.

        Args:
            num_frames: int - The number of frames to cover in the packing string

        Return:
            str - The string to be used with the struct module for unpacking, packing the given number of frames for the
            current audio format described in this object.
        """
        unpack_fmt = '<%i' % ( num_frames * self._n_channels )
        if self._bit_depth == 16:
            unpack_fmt += 'h'
        elif self._bit_depth == 32:
            unpack_fmt += 'i'
        else:
            raise Exception('Unsupporeted bit depth format for packing data.')
        return unpack_fmt
    
    def _read_samples_interleaved_int(self, n=None):
        """
        Reads all samples from the wav file as integers in an interleaved list.
        This replaces any previous data read from the wav file.

        Args:
            n: int - The number of frames (samples per channel) to read.

        Return:
            list[int] - A list of interleaved samples from the audio file.

            int - The number of samples successully read from file. May be less than
            the number of requested samples due to EOF.
        """
        # NOTE [matt.c.mccallum 03.04.19]: We don't use the wave file context-manager below because of problems using it in Python 2.7.10
        if n is None: n = self._num_frames
        data = self._wav_file.readframes( n )
        n_read_frames = int(len(data)*8/self._bit_depth/self._n_channels)
        if (n_read_frames < n): n = n_read_frames

        data = struct.unpack( self._packing_string( n ), data )

        data = list(data)

        return data, n

    def read(self, n=None):
        """
        Reads all samples from the wav file as floats in the range -1.0 <= sample <= 1.0.
        This replaces any previous data read from the wav file.

        Args:
            n: int - The number of samples to read.

        Return:
            np.ndarray - An array of dimensions (num_channels, num_frames) containing float valued audio samples.
        """
        data, n = self._read_samples_interleaved_int( n )
        if n == 0:
            return None

        return_array = np.zeros( ( self._n_channels, n ), dtype=np.float32 )
        for channel in range( self._n_channels ):
            return_array[channel,:] = np.array( data[channel::self._n_channels] )/( 2.0**self._bit_depth )

        data = return_array

        return data

    def __len__(self):
        return self._num_frames

    @property
    def wav_file(self):
        return self._wav_file
    @wav_file.setter
    def wav_file(self, file):
        self._raw_file = file
        self._wav_file = wave.open(file, 'rb')
        self._samp_rate = self._wav_file.getframerate()
        self._n_channels = self._wav_file.getnchannels()
        self._bit_depth = self._wav_file.getsampwidth() * 8
        self._num_frames = min(self._wav_file.getnframes(), int(self._samp_rate*self._max_len))

    @property
    def sample_rate(self):
        return self._sample_rate



# coding=utf-8
# Copyright 2021 Pandora Media, LLC.
#
# Licensed under the GNU GPL License, Version 3.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gnu.org/licenses/gpl-3.0.en.html
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python standard library imports
# None.

# Third party imports
from scooch import Param
from ...source_files import SourceFile

# Local imports
from .. import Feature


class SourceFeature(Feature):
    """
    A feature that is derived directly from raw data, e.g., a data file.
    """

    #
    # SCOOCH Configuration
    #
    _input_file = Param(
        SourceFile,
        doc="The file object defining the parameters of the raw data that this feature is constructed from."
    )

    _CHUNK_SIZE = 44100*60*15

    # 
    # Methods
    #
    def from_file(self, fname):
        """
        Takes a file and processes its data in chunks to form a feature.

        Args:
            fname: str - The path to the input file from which this feature is constructed.
        """
        # Load file
        self._input_file.load(fname)
        
        # Read samples into data
        processed_input_frames = 0
        while processed_input_frames<len(self._input_file):
            data = self._extract(self._input_file, processed_input_frames, self._CHUNK_SIZE)
            processed_input_frames += self._CHUNK_SIZE
            self.add_data(data)

    def clear(self):
        """
        Clears any previously analyzed feature data, ready for a new analysis.
        """
        super().clear()
        self._input_file.close()

    def __len__(self):
        """
        Returns the number of bytes / samples / indices in the input data file.
        """
        return len(self._input_file)



# coding=utf-8
# Copyright 2021 Pandora Media, LLC.
#
# Licensed under the GNU GPL License, Version 3.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gnu.org/licenses/gpl-3.0.en.html
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python standard library imports
# None.

# Third party imports
import tensorflow as tf

# Local imports
# None.


class ScalarMultiply(tf.keras.layers.Layer):
    """
    Implements a very simple layer that multiplies a tensor by a scalar. This scalar
    can be constant or learnable.
    """

    def __init__(self, init_gain, learnable=False, **kwargs):
        """
        **Constructor**

        Args:
            init_gain: <float> - The initial scalar gain that this layer will apply.

            learnable: <bool> - Whether to update the gain during training.
        """
        self._init_gain = init_gain
        self._learnable = learnable

        if self._learnable:
            self.gain = tf.Variable(self._init_gain, trainable=True)
        else:
            self.gain = tf.constant(self._init_gain)

        super(ScalarMultiply, self).__init__(**kwargs)

    def call(self, input):
        """
        Apply the layer to an input.

        Args:
            inputs: <tf.Tensor> - The input tensor to apply the scalar to.

        Return:
            <tf.Tensor> - The output tensor.
        """
        return self.gain*input

    def compute_output_shape(self, input_shape):
        """
        Get the output shape - identical to input.

        Args:
            input_shape: <tf.Tensor> - The input shape.

        Return:
            <tf.Tensor> - The output shape.
        """
        return input_shape.get_shape()

    def get_config(self):
        """
        Get the parameters of the layer for serialization, so that it can be reconstructed
        / deserialized.

        Return:
            <dict(str, object)> - A dictionary containing the properties required to
            reconstruct the object.
        """
        return {'init_gain': self._init_gain, 'learnable': self._learnable}

    @classmethod
    def from_config(cls, cfg):
        """
        Construct the object from previously serialized parameters (does not load weights).

        Args:
            cfg: <dict(str, object)> - The parameters from a serialized layer.

        Return:
            <ScalarMultiply> - The reconstructed ScalarMultiply layer.
        """
        return cls(**cfg)



# coding=utf-8
# Copyright 2024 Pandora Media, LLC.
#
# Licensed under the GNU GPL License, Version 3.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gnu.org/licenses/gpl-3.0.en.html
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Local imports
# None.

# Third party imports
# None.

# Python standard library imports
import setuptools
import platform


long_description = """
# MULE

The Musicset Unsupervised Large Embedding (MULE) module is your 
music-audio workhorse!

This module contains [SCOOCH](https://github.com/PandoraMedia/scooch) configurable code to run a simple 
analysis pipeline to extract audio embeddings from audio files which
may then be used for downstream music understanding purposes.

This module requires FFMpeg to read audio files, which may be 
downloaded [here](https://ffmpeg.org/download.html).

In order to create MULE embeddings, you will need a SCOOCH configuration
describing the pipeline, and the model weights. Both are licensed under 
the [CC BY-NC 4.0](https://creativecommons.org/licenses/by-nc/4.0/legalcode) license, and can be found in this [module's github repository](https://github.com/PandoraMedia/music-audio-representations).

To create embeddings for a single audio file, e.g., `test.wav` in the current
directory, you can use this module in conjunction with the provided configuration
and model weights:

```
pip install sxmp-mule
git clone https://github.com/PandoraMedia/music-audio-representations.git
cd ./music-audio-representations
mule analyze --config ./supporting_data/configs/mule_embedding.yml -i ../test.wav -o ./embedding.npy
```

For more information on this module, please check out the publication:

[*Supervised and Unsupervised Learning of Audio Representations for Music Understanding*](https://arxiv.org/abs/2210.03799), **M. C. McCallum**, F. Korzeniowski, S. Oramas, F. Gouyon, A. F. Ehmann.

"""


# Tensorflow on ARM processors (i.e., macbooks with M1 or M2 or newer processors)
# need a different compilation of tensorflow.
# More details on the correct combinations of tensorflow packages below can be 
# found here:
#   https://developer.apple.com/metal/tensorflow-plugin/
REQUIRED_TF = []
if platform.processor()=='arm':
    REQUIRED_TF += [
        'tensorflow-macos==2.9.1',
        'tensorflow-metal==0.5.0'
    ]
else:
    REQUIRED_TF = ['tensorflow==2.9.1']


REQUIRED_PACKAGES = [
    'numpy',
    'librosa==0.9.2',
    'click==8.1.7',
    'scooch>=1.0.4'
] + REQUIRED_TF


setuptools.setup(
    name='sxmp-mule',
    version='1.1.1',
    description='',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/PandoraMedia/music-audio-representations",
    author="Matt C. McCallum",
    author_email="mmccallum@pandora.com",
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Multimedia :: Sound/Audio :: Analysis',
    ],
    project_urls={
        'Documentation': 'https://github.com/PandoraMedia/music-audio-representations',
        'Bug Reports': 'https://github.com/PandoraMedia/music-audio-representations/issues',
        'Source': 'https://github.com/PandoraMedia/music-audio-representations',
    },
    license='GNU GPL 3.0',
    keywords='mule audio music embeddings machine learning',

    install_requires=REQUIRED_PACKAGES,
    python_requires='>=3.8',
    packages=setuptools.find_packages(),
    include_package_data=True,

    # CLI
    entry_points = {
        'console_scripts': ['mule=mule.cli:main']
    }
)



# coding=utf-8
# Copyright 2021 Pandora Media, LLC.
#
# Licensed under the GNU GPL License, Version 3.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gnu.org/licenses/gpl-3.0.en.html
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Local imports
from .model import Model

# Third party imports
import tensorflow as tf
from scooch import Param

# Python standard library imports
# None.


class MLP(Model):
    """
    A model where all layers form dense transformations interleaved with non-linearities.
    """

    #
    # SCOOCH Configuration
    #
    _layer_widths = Param(
        list,
        doc="The number of neurons in each and every layer of the dense net"
    )
    _activation = Param(
        str,
        default='relu',
        doc="A string specifying the Tensorflow activation function at the output of each and every neuron at each and every layer."
    )
    _l2_regularization = Param(
        float,
        default=None,
        doc="The parameter applied to l2 regularization of dense net weights. If None, no regularization will be applied."
    )
    _input_dropout = Param(
        float,
        default=None,
        doc="Dropout probability applied directly to the input features of the model (None for no dropout)"
    )
    _layer_dropouts = Param(
        list,
        default=None,
        doc="Dropout probability for a layer added after the activation is applied at each hidden layer (None, or the same size as `layer_widths`)"
    )
    _output_dropout = Param(
        float,
        default=None,
        doc="Dropout probability applied directly to the output of the model, after any transformations (None for no dropout)"
    )

    # 
    # Methods
    #
    def _make_model(self):
        """
        Defines a model with an arbitrary number of dense layers with an arbitrary number
        of neurons, and an arbitrary activation function.

        Return:
            tf.keras.Model - A trainable Keras model.
        """
        input = tf.keras.Input(shape=self._input_shapes[0][1:]) # NOTE [matt.c.mccallum 11.29.21]: Assume the first index is the input

        output_size = self._output_shapes[0][1]

        layers = []

        if self._input_dropout:
            layers += [tf.keras.layers.Dropout(self._input_dropout)]

        for idx, layer_width in enumerate(self._layer_widths):
            layers += [tf.keras.layers.Dense(layer_width, activation=self._activation, kernel_regularizer=self._get_regularizer())]
            if self._layer_dropouts and self._layer_dropouts[idx]:
                layers += [tf.keras.layers.Dropout(self._layer_dropouts[idx])]
                
        layers += [tf.keras.layers.Dense(output_size, activation='linear', kernel_regularizer=self._get_regularizer())]
        if self._output_dropout:
            layers += [tf.keras.layers.Dropout(self._output_dropout)]

        output = input
        for lyr in layers:
            output = lyr(output)

        return self._MODEL_CLASS([input], [output])

    def _get_regularizer(self):
        """
        Returns the regularizer according to the object's configuration.

        Return:
            <tf.keras.regularizers.Regularizer or None> - The regularizer to use in model layers.
        """
        if self._l2_regularization:
            return tf.keras.regularizers.L2(l2=self._l2_regularization)
        else:
            return None



# coding=utf-8
# Copyright 2021 Pandora Media, LLC.
#
# Licensed under the GNU GPL License, Version 3.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gnu.org/licenses/gpl-3.0.en.html
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python standard library imports
# None.

# Third party imports
import numpy as np
from scooch import Param
import librosa

# Local imports
from .transform_feature import TransformFeature


class MelSpectrogram(TransformFeature):
    """
    A feature that is a mel spectrogram transformation of its input.
    """

    #
    # SCOOCH Configuration
    #
    _n_fft = Param(
        int,
        default=1024,
        doc="Size of the FFT operation, prior to mel window summarization."
    )
    _hop_length = Param(
        int,
        default=256,
        doc="The number of samples between the start of any two consecutive analysis windows inclusive of the sample at the start of the first analysis window."
    )
    _win_length = Param(
        int,
        default=512,
        doc="Number of audio samples in each window"
    )
    _window = Param(
        str,
        default='hamming',
        doc="A string selecting the window type"
    )
    _n_mels = Param(
        int,
        default=96,
        doc="Number of mel windows each spectrogram window is summarized into across frequency"
    )
    _fmin = Param(
        float,
        default=40.0,
        doc="The frequency in Hz of the lowest frequency mel band"
    )
    _fmax = Param(
        float,
        default=8000.0,
        doc="The frequency in Hz of the highest frequeny mel band"
    )
    _norm = Param(
        float,
        default=2.0,
        doc="The order of normalization (e.g., 2.0 is power normalization)"
    )
    _mag_compression = Param(
        str,
        default='linear',
        doc="The type of compression to be applied to the magnitude mel values"
    )
    _mag_range = Param(
        float,
        default=9999999999.0,
        doc="The range below the maximum magnitude to limit the magnitudes to. If set to None, the value range will not be limited."
    )
    _power = Param(
        float,
        default=2.0,
        doc="The exponent to be used to compute the power spectrogram"
    )
    _htk = Param(
        bool,
        default=False,
        doc="Whether or not to use the HTK formula for mel frequencies (see librosa docs)"
    )
    _pad_mode = Param(
        str,
        default='reflect',
        doc="How to pad the signal in the beginning and at the end (see np.pad)"
    )

    _COMPRESSION_FUNCS = {
        '10log10': lambda x: 10*np.log10(x),
        'log10_nonneg': lambda x: np.log10(10000.0*x + 1.0),
        'log': lambda x: np.log(x),
        'linear': lambda x: x,
        None: lambda x: x,
    }

    _SMALLEST_MAGNITUDE = -9999999999.0

    #
    # Methods
    #
    def _extract(self, source_feature, start_time, chunk_size):
        """
        Extracts a mel spectrogram of the input feature over the provided time range.

        Args:
            source_feature: mule.features.AudioWaveform - The feature to transform into a mel spectrogram

            start_time: int - The index in the feature at which to start extracting / transforming.

            chunk_size: int - The length of the chunk following the `start_time` to extract
            the feature from.
        """
        # TODO [matt.c.mccallum 11.19.22]: Better handling of boundary conditions by zero padding only start and end of input, not of each chunk.

        data = np.squeeze(super()._extract(source_feature, start_time, chunk_size))
        
        mel_data = librosa.feature.melspectrogram(
            y=data, 
            sr=source_feature.sample_rate, 
            n_fft=self._n_fft, 
            hop_length=self._hop_length, 
            win_length=self._win_length, 
            window=self._window, 
            center=True, 
            pad_mode=self._pad_mode, 
            power=self._power,
            n_mels=self._n_mels,
            fmin=self._fmin,
            fmax=self._fmax,
            norm=self._norm,
            htk=self._htk,
            dtype=np.float32
        )

        mel_data = self._COMPRESSION_FUNCS[self._mag_compression](mel_data)
        mel_data = np.nan_to_num(mel_data, nan=self._SMALLEST_MAGNITUDE, posinf=self._SMALLEST_MAGNITUDE, neginf=self._SMALLEST_MAGNITUDE)

        if self._mag_range is not None:
            max_val = np.amax(mel_data)
            mel_data = mel_data - max_val
            mel_data = np.maximum(mel_data, -self._mag_range)
        
        return mel_data


# coding=utf-8
# Copyright 2021 Pandora Media, LLC.
#
# Licensed under the GNU GPL License, Version 3.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gnu.org/licenses/gpl-3.0.en.html
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python standard library imports
from abc import abstractmethod

# Third party imports
from scooch import Configurable
from scooch import ConfigList
from scooch import Param
import tensorflow as tf

# Local imports
from .losses import Loss
from .optimizers import Optimizer
from .metrics import TrainMetric


class Model(Configurable):
    """
    A class defining the architecture of a model along with the training parameters such as loss function and optimizer.
    """

    # 
    # SCOOCH Configuration
    #
    _optimizer = Param(
        Optimizer,
        doc="A configuration for a Tensorflow optimizer to use for model training."
    )
    _loss = Param(
        Loss,
        default=None,
        doc="A SCOOCH configuration / dictionary specifying the loss function used during training."
    )
    _metrics = Param(
        ConfigList(TrainMetric),
        doc=" - A list of metric configurations to be compiled into the model, to observe whilst training."
    )

    _MODEL_CLASS = tf.keras.Model

    # 
    # Methods
    #
    def __init__(self, cfg):
        """
        **Constructor:**

        Args:
            cfg - A SCOOCH Config object defining the configuration parameters for this class.
        """
        super().__init__(cfg)
        self._model = None
        self._input_shapes = None
        self._output_shapes = None

    def init(self, input_shapes, output_shapes):
        """
        Peform any actions immediately after the model is defined. For example, specify IO shapes which
        may affect model architecture, define the loss and compile the model.

        Args:
            input_shapes: <list(list(list(int)))> - A list containing lists of dimensions of all input
            tensors.

            output_shapes: <list(list(list(int)))> - A list containing lists of dimensions of all output
            tensors.
        """
        self._input_shapes = input_shapes
        self._output_shapes = output_shapes

        strategy = tf.distribute.MirroredStrategy()

        with strategy.scope():

            self._model = self._make_model()

            # Initialize these here, now that we're in the distribution strategy scope
            self._loss.init()
            self._optimizer.init()
            if self._metrics is not None:
                for mt in self._metrics:
                    if isinstance(mt, TrainMetric):
                        mt.init()

            # Compile the model
            self._model.compile(optimizer=self._optimizer, loss=self._loss, metrics=self._metrics)
    
    @abstractmethod
    def _make_model(self):
        """
        The method where the actual model definition occurs. This is to be overridden by each of the
        model classes, defining what the model architecture is.

        Return:
            <tf.keras.Model> - The model instance that this class configurizes and encapsulates, for training,
            inference, or otherwise.
        """
        raise NotImplementedError(f'This model, {self.__class__.__name__}, has not been implemented')

    @property
    def model(self):
        """
        Type: <tf.keras.Model>
            The keras model instance that is encapsulated within this configurable model wrapper.
        """
        return self._model



# coding=utf-8
# Copyright 2021 Pandora Media, LLC.
#
# Licensed under the GNU GPL License, Version 3.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gnu.org/licenses/gpl-3.0.en.html
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python standard library imports
# None.

# Third party imports
from scooch import Configurable

# Local imports
# None.


class SourceFile(Configurable):
    """
    Base class for SCOOCH configurable file readers.
    """

    def load(self, fname):
        """
        Any preprocessing steps to load a file prior to reading it.

        Args:
            fname: file-like - A file like object to be loaded.
        """
        raise NotImplementedError(f"The class, {self.__class__.__name__}, has no method for loading files")

    def read(self, n):
        """
        Reads an amount of data from the file.

        Args:
            n: int - A size parameter indicating the amount of data to read.

        Return:
            object - The decoded data read and in memory.
        """
        raise NotImplementedError(f"The class, {self.__class__.__name__}, has no method for reading files")

    def close(self):
        """
        Closes any previously loaded file.
        """
        raise NotImplementedError(f"The class, {self.__class__.__name__}, has no method for closing files")

    def __len__(self):
        raise NotImplementedError(f"The class, {self.__class__.__name__} has no method for determining file data length")
        


# coding=utf-8
# Copyright 2024 Pandora Media, LLC.
#
# Licensed under the GNU GPL License, Version 3.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gnu.org/licenses/gpl-3.0.en.html
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .analysis import Analysis
